/**
 * CMI : Cluster Method Invocation
 * Copyright (C) 2008-2009 Bull S.A.S.
 * Contact: cmi@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * --------------------------------------------------------------------------
 * $Id: CMIConfigurator.java 2438 2010-02-24 16:20:30Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.cmi.config;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ow2.cmi.controller.factory.ClusterViewManagerFactory;
import org.ow2.cmi.controller.factory.ClusterViewManagerFactoryConfig;
import org.ow2.cmi.reference.ServerRef;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.xml.DocumentParser;
import org.ow2.util.xml.XMLUtils;
import org.ow2.util.xmlconfig.XMLConfiguration;
import org.ow2.util.xmlconfig.XMLConfigurationException;
import org.ow2.util.xmlconfig.properties.SystemPropertyResolver;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Utility class to create a CMI configuration.
 * @author Loris Bouzonnet
 */
public final class CMIConfigurator {

    /**
     * Default configuration filename.
     */
    public static final Object DEFAULT_FILENAME = "cmi-jgroups-config.xml";

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(CMIConfigurator.class);

    /**
     * Utility class: no public constructor.
     */
    private CMIConfigurator() {
    }

    /**
     * @param environment an environment to retrieve the CMI properties
     * @param onlyJndi true indicate if only JNDI (and not the manager) must be configured
//     * @return The CMI configuration.
     * @throws CMIConfigException any.
     */
    public static CMIConfig<?> findAndExtractCMIConfig(
            final Hashtable<?, ?> environment, final boolean onlyJndi) throws CMIConfigException {

        // 1. Extract the CMI configuration from the given environment
        CMIConfig<?> cmiConfig = extractCMIConfig(environment, onlyJndi);

        if(cmiConfig == null) {
            // environment doesn't contain a configuration: trying with a XML file
            Object filename =
                environment.get(CMIProperty.CONF_FILENAME.getPropertyName());
            if(filename == null) {
                filename =
                    System.getProperty(CMIProperty.CONF_FILENAME.getPropertyName());
                logger.debug("File name retrievd from " + filename + " sytem properties.");
            }
            if(filename == null) {
                filename = DEFAULT_FILENAME;
            }
            URL url;
            try {
                url = getUrlForFilename(environment, (String) filename);
            } catch (MalformedURLException e) {
                logger.error("Cannot configure the cluster view manager factory from the file: " + filename, e);
                throw new CMIConfigException(
                        "Cannot configure the cluster view manager factory from the file: " + filename, e);
            }
            cmiConfig = CMIConfigurator.extractCMIConfig(url);
        }
        return cmiConfig;
    }

    /**
     * @param environment an environment to retrieve the CMI properties
     * @param onlyJndi true indicate if only JNDI (and not the manager) must be configured
     * @return The CMI configuration.
     * @throws CMIConfigException any.
     */
    public static CMIConfig<?> extractCMIConfig(
            final Hashtable<?, ?> environment, final boolean onlyJndi) throws CMIConfigException {

        CMIConfig<?> cmiConfig;
        boolean serverModeEnabled = isServerModeEnabled(environment);
        Object  managerClassname = environment.get(CMIProperty.REPLICATION_MANAGER_CLASS.getPropertyName());

        if(managerClassname != null) {
            serverModeEnabled = true;
        }

        // Set the JNDI properties
        JNDIConfig jndiConfig = extractJNDIConfig(environment);

        // Only continue if the JNDI is not null or if a manager must be created by using the given environment
        if(jndiConfig == null && (!serverModeEnabled || managerClassname == null || onlyJndi)) {
            return null;
        }

        if(onlyJndi) {
            cmiConfig = new CMIConfig<ClusterViewManagerFactoryConfig<?>>();
            cmiConfig.setJNDIConfig(jndiConfig);
            return cmiConfig;
        }

        ClusterViewManagerFactoryConfig<?> clusterViewManagerFactoryConfig;

        if(serverModeEnabled) {
            CMIConfig<org.ow2.cmi.controller.factory.server.ClusterViewManagerFactoryConfig> cmiConfigServer =
                new CMIConfig<org.ow2.cmi.controller.factory.server.ClusterViewManagerFactoryConfig>();
            org.ow2.cmi.controller.factory.server.ClusterViewManagerFactoryConfig serverClusterViewManagerFactoryConfig =
                new org.ow2.cmi.controller.factory.server.ClusterViewManagerFactoryConfig();
            if(managerClassname == null) {
                managerClassname = org.ow2.cmi.controller.factory.server.ClusterViewManagerFactoryConfig.DEFAULT_MANAGER;
            }
            serverClusterViewManagerFactoryConfig.setServerClusterViewManagerClassname(
                    (String) managerClassname);
            clusterViewManagerFactoryConfig = serverClusterViewManagerFactoryConfig;
            cmiConfigServer.setClusterViewManagerFactoryConfig(serverClusterViewManagerFactoryConfig);
            cmiConfig = cmiConfigServer;

            Object filename = System.getProperty(CMIProperty.CONF_FILENAME.getPropertyName());

            if(filename == null) {
                filename =
                    environment.get(CMIProperty.CONF_FILENAME.getPropertyName());
            }
            if(filename == null) {
                filename = DEFAULT_FILENAME;
            }
            URL url;
            try {
                url = getUrlForFilename(environment, (String) filename);
            } catch (MalformedURLException e) {
                logger.error("Cannot configure the cluster view manager factory from the file: " + filename, e);
                throw new CMIConfigException(
                        "Cannot configure the cluster view manager factory from the file: " + filename, e);
            }
            XMLConfiguration xmlConfiguration = new XMLConfiguration(url, "cmi-config-mapping.xml");

            // Resolve properties as system properties
            xmlConfiguration.setPropertyResolver(new SystemPropertyResolver());

            try {
                xmlConfiguration.configure(clusterViewManagerFactoryConfig);
            } catch (XMLConfigurationException e) {
                logger.error("Cannot configure CMI", e);
                throw new CMIConfigException("Cannot configure CMI", e);
            }
        } else {
            CMIConfig<org.ow2.cmi.controller.factory.client.ClusterViewManagerFactoryConfig> cmiConfigClient =
                new CMIConfig<org.ow2.cmi.controller.factory.client.ClusterViewManagerFactoryConfig>();
            org.ow2.cmi.controller.factory.client.ClusterViewManagerFactoryConfig clientClusterViewManagerFactoryConfig =
                new org.ow2.cmi.controller.factory.client.ClusterViewManagerFactoryConfig();
            clusterViewManagerFactoryConfig = clientClusterViewManagerFactoryConfig;
            cmiConfigClient.setClusterViewManagerFactoryConfig(clientClusterViewManagerFactoryConfig);
            cmiConfig = cmiConfigClient;
            org.ow2.cmi.controller.client.Config clientConfig = new org.ow2.cmi.controller.client.Config();
            clientClusterViewManagerFactoryConfig.setConfig(clientConfig);
        }
        clusterViewManagerFactoryConfig.setAutoStarted(extractAutoStart(environment));

        cmiConfig.setJNDIConfig(jndiConfig);

        return cmiConfig;
    }

    /**
     * Extract CMI configuration from given URL pointing to a xml file.
     * @param url the given url.
     * @return The CMI configuration.
     * @throws CMIConfigException any.
     */
    public static CMIConfig<?> extractCMIConfig(final URL url) throws CMIConfigException {
        Document document;
        InputStream inputStream = null;
        try {
            URLConnection urlConnection = url.openConnection();
            urlConnection.setUseCaches(false);
            inputStream = urlConnection.getInputStream();
            document = DocumentParser.getDocument(inputStream, false, null);
            logger.info("Configuring CMI from file: {0}", url.getPath());
        } catch (Exception e) {
            throw new CMIConfigException("Cannot get a document on the given url '" + url + "'.", e);
        } finally {
            if(inputStream != null) {
                try {
                    inputStream.close();
                } catch (IOException e) {
                    logger.debug("Unable to close the input stream", e);
                }
            }
        }
        Element element = document.getDocumentElement();
        NodeList nodeList = element.getElementsByTagName("server");
        int nbNodes = nodeList.getLength();
        if(nbNodes > 1) {
            logger.error("Invalid XML configuration file: {0} nodes with name 'server' found.", nbNodes);
            throw new CMIConfigException(
                    "Invalid XML configuration file: "
                    + nbNodes + " nodes with name 'server' found.");
        }
        CMIConfig<?> cmiConfig;
        if(nbNodes == 1) {
            cmiConfig = new CMIConfig<org.ow2.cmi.controller.factory.server.ClusterViewManagerFactoryConfig>();
            element = (Element) nodeList.item(0);
            logger.info("Server configuration found");
        } else {
            nodeList = element.getElementsByTagName("client");
            nbNodes = nodeList.getLength();
            if(nbNodes != 1) {
                logger.error("Invalid XML configuration file: {0} nodes with name 'client' found.", nbNodes);
                throw new CMIConfigException(
                        "Invalid XML configuration file: "
                        + nbNodes + " nodes with name 'client' found.");
            }
            cmiConfig = new CMIConfig<org.ow2.cmi.controller.factory.client.ClusterViewManagerFactoryConfig>();
            element = (Element) nodeList.item(0);
            String noCmi = XMLUtils.getAttributeValue(element, "noCmi");
            if(Boolean.parseBoolean(noCmi)) {
                cmiConfig.setNoCmi(true);
            }
            logger.info("Client configuration found");
        }
        configure(url, element, cmiConfig);
        if (System.getProperty(CMIProperty.CONF_FILENAME.getPropertyName()) == null) {
            String f = url.getFile();
            int ind  = f.lastIndexOf("/");
            String fileName = f.substring(ind + 1);
            System.setProperty(CMIProperty.CONF_FILENAME.getPropertyName(), fileName);
            logger.debug("Configuration " + f + "is set in system property");
        }
        return cmiConfig;
    }

    /**
     * Returns the initial provider URLs by extracting it from environment.
     * @param protocol The protocol to extract server references from.
     * @param providerURLs Set of provider URLs
     * @return a list of references on server found in environment
     * @throws CMIConfigException if environment doesn't contain provider URL or if the property is malformed
     */
    public static List<ServerRef> extractServerRefs(final String protocol, final List<String> providerURLs)
    throws CMIConfigException {
        if(protocol == null) {
            throw new IllegalArgumentException("Protocol missing");
        }
        List<ServerRef> serverRefs;
        int size = providerURLs.size();

        serverRefs = new ArrayList<ServerRef>(size);
        for(String providerURL : providerURLs) {
            try {
                serverRefs.add(new ServerRef(protocol, providerURL));
            } catch (MalformedURLException e) {
                logger.error("The following provider URL is malformed {0}", providerURL, e);
                throw new CMIConfigException("The following provider URL is malformed " + providerURL, e);
            } catch (UnknownHostException e) {
                logger.error("The host name of the provider URL {0} cannot be resolved", providerURL, e);
                throw new CMIConfigException(
                        "The host name of the provider URL " + providerURL + " cannot be resolved", e);
            }
        }
        return serverRefs;
    }

    /**
     * Apply a configuration to a given cluster view manager factory.
     * @param clusterViewManagerFactory The manager factory to use.
     * @param cmiConfig The CMI configuration
     * @throws CMIConfigException any.
     */
    public static void configure(
            final ClusterViewManagerFactory clusterViewManagerFactory, final CMIConfig<?> cmiConfig) throws CMIConfigException {
        ClusterViewManagerFactoryConfig<?> clusterViewManagerFactoryConfig =
            cmiConfig.getClusterViewManagerFactoryConfig();
        // Check it
        if(clusterViewManagerFactoryConfig == null) {
            logger.error("Cluster view manager config missing!");
            throw new CMIConfigException("Cluster view manager config missing!");
        }

        if(clusterViewManagerFactoryConfig instanceof org.ow2.cmi.controller.factory.client.ClusterViewManagerFactoryConfig) {
            JNDIConfig jndiConfig = cmiConfig.getJNDIConfig();
            // Check it
            if(jndiConfig == null) {
                logger.error("JNDI config missing!");
                throw new CMIConfigException("JNDI config missing!");
            }
            Set<ProtocolConfig> protocolConfigs = jndiConfig.getProtocolConfigs();
            // Check it
            if(protocolConfigs == null) {
                logger.error("Protocol config missing!");
                throw new CMIConfigException("Protocol config missing!");
            }
            ProtocolConfig currentProtocolConfig = getCurrentProtocolConfig(jndiConfig);
            String protocol = JNDIConfig.getCurrentProtocolName();

            // Check it
            if(currentProtocolConfig == null) {
                logger.error("Missing configuration for " + protocol);
                throw new CMIConfigException("Missing configuration for " + protocol);
            }
            String wrappedFactoryName = currentProtocolConfig.getInitialContextFactoryName();
            ProviderUrls providerUrls = currentProtocolConfig.getProviderUrls();
            if(providerUrls == null) {
                logger.error("Missing configuration for " + protocol);
                throw new CMIConfigException("Missing configuration for " + protocol);
            }
            List<String> initialProviderURLs = providerUrls.getProviderUrls();
            String localRegistryUrl = currentProtocolConfig.getLocalRegistryUrl();

            ((org.ow2.cmi.controller.factory.client.ClusterViewManagerFactoryConfig) clusterViewManagerFactoryConfig)
                .setClusterViewProviderUrls(initialProviderURLs);
            ((org.ow2.cmi.controller.factory.client.ClusterViewManagerFactoryConfig) clusterViewManagerFactoryConfig)
                .setInitialContextFactoryName(wrappedFactoryName);
            ((org.ow2.cmi.controller.factory.client.ClusterViewManagerFactoryConfig) clusterViewManagerFactoryConfig)
                .setProtocol(protocol);
            ((org.ow2.cmi.controller.factory.client.ClusterViewManagerFactoryConfig) clusterViewManagerFactoryConfig)
                .setLocalRegistryUrl(localRegistryUrl);
        } else {
            String protocol = JNDIConfig.getCurrentProtocolName();
            if(protocol != null) {
                patchProtocol(protocol);
            }
        }
        String confDir = System.getProperty(CMIProperty.CONF_DIR.getPropertyName());
        if (confDir != null) {
            clusterViewManagerFactoryConfig.getConfig().setConfDir(confDir);
        }
        clusterViewManagerFactory.configure(cmiConfig);
    }


    public static ProtocolConfig getCurrentProtocolConfig(final JNDIConfig jndiConfig) throws CMIConfigException {

        ProtocolConfig currentProtocolConfig = null;
        Set<ProtocolConfig> protocolConfigs = jndiConfig.getProtocolConfigs();
        String protocol = JNDIConfig.getCurrentProtocolName();

        if(protocolConfigs == null) {
            logger.error("Protocol configs missing!");
            throw new CMIConfigException("Protocol configs missing!");
        }

        if(protocol == null) {
            if(protocolConfigs.size() != 1) {
                logger.error("Protocol name missing!");
                throw new CMIConfigException("Protocol name missing!");
            }
            currentProtocolConfig = (ProtocolConfig) protocolConfigs.toArray()[0];
            protocol = currentProtocolConfig.getName();
            JNDIConfig.setCurrentProtocolName(protocol);
            logger.debug("No default protocol specified, by default using the protocol {0}", protocol);
        } else {
            for(ProtocolConfig protocolConfig : protocolConfigs) {
                if(protocolConfig.getName().equals(protocol)) {
                    currentProtocolConfig = protocolConfig;
                    break;
                }
            }
        }
        return currentProtocolConfig;
    }

    /**
     * Gets URL for a given file located in the current configuration directory.
     * @param environment The environment to get the configuration directory from
     * @param filename the file to get the URL from
     * @return The URL.
     * @throws MalformedURLException if the URL is malformed.
     */
    private static URL getUrlForFilename(final Hashtable<?, ?> environment, final String filename) throws MalformedURLException {
        // 1. Try to retrieve it from a file specified in the given environment property 'cmi.conf.dir' and 'cmi.conf.filename'
        Object confURL = environment.get(CMIProperty.CONF_DIR.getPropertyName());

        if(confURL == null) {
            // 2. Try to retrieve it from a file specified in the system property 'cmi.conf.url'
            confURL = System.getProperty(CMIProperty.CONF_DIR.getPropertyName());
        }
        if(confURL == null) {
            logger.warn("No configuration directory set: using the classpath");
            // 3. Use the classpath
            URL ret = CMIConfigurator.class.getClassLoader().getResource(filename);
            if (ret == null) {
                ret = Thread.currentThread().getContextClassLoader().getResource(filename);
            }
            return ret;
        } else {
            // 2.|3. Use the directory
            return new File((String) confURL, filename).toURI().toURL();
        }
    }

    /**
     * @param url
     * @param element
     * @param cmiConfig
     * @throws CMIConfigException
     */
    private static void configure(final URL url, final Element element, final Object cmiConfig) throws CMIConfigException {
        long tStart = System.currentTimeMillis();
        logger.debug("Starting configuration of CMI");
        XMLConfiguration xmlConfiguration = new XMLConfiguration(url, "cmi-config-mapping.xml");

        // Resolve properties as system properties
        xmlConfiguration.setPropertyResolver(new SystemPropertyResolver());

        try {
            xmlConfiguration.configure(cmiConfig, element, null);
        } catch (XMLConfigurationException e) {
            logger.error("Cannot configure CMI", e);
            throw new CMIConfigException("Cannot configure CMI", e);
        }
        if (logger.isInfoEnabled()) {
            logger.debug("Configuration done in : " + (System.currentTimeMillis() - tStart) + " ms");
        }
    }

    /**
     * Patches the protocol by adding adequate prodelegate.
     * @param protocol the protocol to patch.
     */
    private static void patchProtocol(final String protocol) {
        String proDelegate = System.getProperty("javax.rmi.CORBA.PortableRemoteObjectClass");
        logger.debug("Current PRO delegate is {0}", proDelegate);
        if(proDelegate == null || proDelegate != null
                && proDelegate.equals("com.sun.corba.se.impl.javax.rmi.PortableRemoteObject")) {
            logger.debug("Using the class 'org.ow2.cmi.rmi.DefaultCMIPRODelegate' as PRO delegate");
            System.setProperty("javax.rmi.CORBA.PortableRemoteObjectClass",
                "org.ow2.cmi.rmi.DefaultCMIPRODelegate");
            if(protocol == "iiop") {
                logger.debug("Using the class 'com.sun.corba.se.impl.presentation.rmi.StubFactoryFactoryStaticImpl'"
                        +"as ORBDynamicStubFactoryFactory");
                System.setProperty("com.sun.CORBA.ORBDynamicStubFactoryFactoryClass",
                    "com.sun.corba.se.impl.presentation.rmi.StubFactoryFactoryStaticImpl");
            }
        }
    }

    /**
     * Extracts JNDI configuration from a given environment.
     * @param environment the environment to extract JNDI configuration from.
     * @return The JNDI configuration.
     * @throws CMIConfigException any.
     */
    private static JNDIConfig extractJNDIConfig(final Hashtable<?, ?> environment) throws CMIConfigException {
        Object protocol =
            environment.get(CMIProperty.WRAPPED_PROTOCOL.getPropertyName());
        // Check
        if(protocol == null) {
            return null;
        }
        JNDIConfig jndiConfig = new JNDIConfig();
        ProtocolConfig protocolConfig = new ProtocolConfig();
        jndiConfig.addProtocolConfig(protocolConfig);
        protocolConfig.setName((String) protocol);
        Object initialContextFactory =
            environment.get(CMIProperty.WRAPPED_INITIAL_CONTEXT_FACTORY.getPropertyName());
        if(initialContextFactory != null) {
            protocolConfig.setInitialContextFactoryName((String) initialContextFactory);
        }
        Object localRegistryUrl =
            environment.get(CMIProperty.LOCAL_REGISTRY_URL.getPropertyName());
        if(localRegistryUrl != null) {
            protocolConfig.setLocalRegistryUrl((String) localRegistryUrl);
        }
        Object providerURLs =
            environment.get(CMIProperty.CLUSTER_VIEW_PROVIDER_URLS.getPropertyName());
        if(providerURLs != null) {
            ProviderUrls pUrls = new ProviderUrls();
            protocolConfig.setProviderUrls(pUrls);
            try {
                pUrls.setProviderUrls(ProviderURLsParser.parse((String) providerURLs));
            } catch (MalformedURLException e) {
                logger.error(
                        "Cannot parse the following urls: {0}.", providerURLs, e);
                throw new CMIConfigException(
                        "Cannot parse the following urls: " + providerURLs, e);
            }
        }
        JNDIConfig.setCurrentProtocolName((String) protocol);
        return jndiConfig;
    }

    /**
     * Checks if server mode is enabled.
     * @param bootstrapEnv The bootstrap environment.
     * @return true if server mode si enabled.
     */
    private static boolean isServerModeEnabled(final Map<?, ?> bootstrapEnv) {
        Object obj = bootstrapEnv.get(CMIProperty.SERVER_MODE.getPropertyName());
        if(obj == null) {
            return false;
        }
        return Boolean.parseBoolean(obj.toString());
    }

    /**
     * Check is the manager is started on creation of the first context.
     * @param bootstrapEnv The bootstrap environment.
     * @return true if so.
     */
    private static boolean extractAutoStart(final Map<?, ?> bootstrapEnv) {
        Object obj = bootstrapEnv.get(CMIProperty.AUTOSTART.getPropertyName());
        if(obj == null) {
            return true;
        }
        return Boolean.parseBoolean(obj.toString());
    }

}
